@@ -2,6 +2,8 @@ import argparse
 import ast
 import os
 import json
+import shlex
+import sys
 import textwrap
 import threading
 import time
@@ -19,6 +21,7 @@ import requests
 
 DEFAULT_SERVER_PORT = 43234 #This is an arbitrary port I chose because its easy to remember and didn't conflict with any known services I could find on https://en.wikipedia.org/wiki/List_of_TCP_and_UDP_port_numbers
 DEFAULT_DELEGATION_SERVER_PORT = 33234 #The port for delegation servers is different, making them easy to find if we have many servers
+DEFAULT_CLUSTER_SESSION_NAME = "Webeval Cluster"
 
 #RP's Web Evaluator
 #    This module provides duct-tape to connect python between computers
@@ -394,7 +397,8 @@ def run_server(server_port:int=None,
                scope:dict=None,
                *,
                sync=True,
-               handler_base_class=SimpleHTTPRequestHandler):
+               handler_base_class=SimpleHTTPRequestHandler,
+               new_thread=False):
     """
     Runs a web_evaluator server, which provides an HTTP interface for remote Python code execution (and can double as a fileserver as well)
 
@@ -412,9 +416,15 @@ def run_server(server_port:int=None,
             accepting code execution requests. This is useful for serving HTML/JS files that can interact with the Python backend.
             If `BaseHTTPRequestHandler`, the server will only support code execution requests and will not serve files.
             You can also provide your own custom handler class to extend the server's functionality.
+            
+        new_thread (bool): If True, starts the server in a new thread and returns an object with a .stop() method.
+            If False (default), blocks indefinitely to handle incoming HTTP requests.
 
     Returns:
-        None. The function will block indefinitely to handle incoming HTTP requests until the process is terminated.
+        If new_thread is False:
+            None. The function will block indefinitely to handle incoming HTTP requests until the process is terminated.
+        If new_thread is True:
+            An object with a .stop() method to stop the server.
 
     The server supports several types of requests, including:
         - Python code execution requests (POST to /webeval/py2py)
@@ -522,12 +532,26 @@ def run_server(server_port:int=None,
     )
     print("Server started at http://%s:%s" % (rp.get_my_local_ip_address(), server_port))
 
-    try:
-        webServer.serve_forever()
-    finally:
-        webServer.server_close()
-
-    print("Server stopped.")
+    if new_thread:
+        server_thread = threading.Thread(target=webServer.serve_forever)
+        server_thread.daemon = True
+        server_thread.start()
+        
+        def stop(self):
+            webServer.shutdown()
+            webServer.server_close()
+            if server_thread.is_alive():
+                server_thread.join()
+            print("Server stopped.")
+        
+        return type('StoppableWebEvaluatorServer', (), {'stop': stop})()
+    else:
+        try:
+            webServer.serve_forever()
+        finally:
+            webServer.server_close()
+        
+        print("Server stopped.")
 
 class Client:
     DEFAULT_PORT = DEFAULT_SERVER_PORT
@@ -743,6 +767,22 @@ class Client:
         assert isinstance(output, Client)
         return output
 
+    @classmethod
+    def from_client(cls, other_client, **kwargs):
+        """
+        Create a new Client based on another Client, with optional parameter overrides.
+        
+        Args:
+            other_client (Client): The client to copy settings from
+            **kwargs: Parameters to override (server_name, server_port, sync, timeout)
+        
+        Returns:
+            Client: A new Client instance with copied settings and overrides applied
+        """
+        client_kwargs = rp.gather_attrs(other_client, 'server_name server_port sync timeout', as_dict=True)
+        client_kwargs.update(kwargs)
+        return cls(**client_kwargs)
+
     def __hash__(self):
         return hash(repr(self))
 
@@ -982,6 +1022,27 @@ class ClientDelegator:
 
         self._update()
 
+        try:
+
+            return self._evaluate_client(client, code, **vars)
+
+        finally:
+            client._busy_count-=1
+
+            #TODO: Figure out why this below assertion sometimes triggers. Makes no sense to me. For now I'll ignore it and just make a warning.
+            #assert client._busy_count>=0
+            if client._busy_count<0:
+                rp.fansi_print("rp.web_evaluator.ClientDelegator.evaluate: client._busy_count<0 - this shouldn't be possible. Setting it to 0. Todo: Investigate this.", 'red', 'bold')
+                client.busy_count=0
+
+            self._update()
+
+    def _evaluate_client(self, client, code, options=None, **vars):
+
+        default_options = dict(abandon_on_unregister=False)
+        options = options or {}
+        options = rp.as_easydict(rp.merged_dicts(default_options, options))
+
         try:
             result = client.evaluate(code, **vars)
             return result
@@ -992,7 +1053,11 @@ class ClientDelegator:
             if self.on_connection_error == 'unregister':
                 #Delete dead clients
                 self.unregister_client(client)
-                return self.evaluate(code, **vars)
+                if not options.abandon_on_unregister:
+                    return self.evaluate(code, **vars)
+                else:
+                    #Used for evaluate_all
+                    return None
             elif self.on_connection_error == 'raise':
                 #Throw an error right away - handle it elsewhere
                 raise
@@ -1005,18 +1070,59 @@ class ClientDelegator:
                         time.sleep(1)
             else:
                 assert False, 'Invalid self.on_connection_error value: '+str(self.on_connection_error)
-
-        finally:
-            client._busy_count-=1
-
-            #TODO: Figure out why this below assertion sometimes triggers. Makes no sense to me. For now I'll ignore it and just make a warning.
-            #assert client._busy_count>=0
-            if client._busy_count<0:
-                rp.fansi_print("rp.web_evaluator.ClientDelegator.evaluate: client._busy_count<0 - this shouldn't be possible. Setting it to 0. Todo: Investigate this.", 'red', 'bold')
-                client.busy_count=0
-
-            self._update()
         
+    def evaluate_all(self, code, **vars):
+        """
+        Submits a job to be executed on all registered clients simultaneously.
+        """
+        def evaluate_on_client(client):
+            return self._evaluate_client(client, code, dict(abandon_on_unregister=True), **vars)
+        
+        output = rp.par_map(evaluate_on_client, self._clients, num_threads=len(self._clients))
+        return [x for x in output if x is not None]
+
+    def _rerank(self):
+        #TODO: This is a work in progress - don't use it yet 
+        results = self.evaluate_all(
+            rp.unindent(
+                """
+                %return _out
+                if 'WEBEVAL_RANK' in vars():
+                    _out = WEBEVAL_RANK
+                else:
+                    _unrank = -__import__("secrets").randbits(1000)-1
+                    _out = _unrank
+                """
+            )
+        )
+
+        values = [x.value for x in results if not x.errored]
+
+        ranks = [x for x in values if x>=0]
+        unranks = [x for x in values if x<0]
+
+        num_ranks = len(values)
+        missing_ranks = sorted(set(range(num_ranks))-set(ranks))
+
+        results = self.evaluate_all(
+            rp.unindent(
+                """
+                %return WEBEVAL_RANK
+
+                if 'WEBEVAL_RANK' not in vars():
+                    _ranks   = [x for x in _values if x >= 0]
+                    _unranks = [x for x in _values if x <  0]
+                    _num_ranks = len(_values)
+                    _missing_ranks = sorted(set(range(_num_ranks)) - set(_ranks))
+                    WEBEVAL_RANK = _missing_ranks[_unranks.index(_unrank)]
+
+                print("WEBEVAL_RANK = {WEBEVAL_RANK}")
+                """
+            ),
+            _values = values,
+        )
+
+
 class ClientRoster:
     """
     The ClientRoster class manages a roster of clients for the ClientDelegator.
@@ -1159,8 +1265,8 @@ def run_delegation_server(server_port=None,
                           *,
                           delegator=None,
                           roster=None,
-                          refresh_interval = 5
-                          ):
+                          refresh_interval = 5,
+                          new_thread=False):
 
     """
     Starts a delegation server that manages a pool of web_evaluator servers for distributed computation.
@@ -1188,8 +1294,8 @@ def run_delegation_server(server_port=None,
     """
 
     #If not specified, create the default versions
-    if roster    is None: roster    = ClientRoster()
-    if delegator is None: delegator = ClientDelegator()
+    if roster      is None: roster      = ClientRoster()
+    if delegator   is None: delegator   = ClientDelegator()
     if server_port is None: server_port = DEFAULT_DELEGATION_SERVER_PORT
 
     if rp.get_port_is_taken(server_port):
@@ -1223,7 +1329,7 @@ def run_delegation_server(server_port=None,
             delegator = delegator,
         )
                 
-        run_server(server_port, scope=scope, sync=False)
+        run_server(server_port, scope=scope, sync=False, new_thread=new_thread)
 
     finally:
         #Exiting the server - kill loose threads
@@ -1247,6 +1353,37 @@ class DelegationClient(Client):
     TODO: Make this more efficient - see run_delegation_server's TODO. As a blackbox, the signatures can stay the same
     """
     DEFAULT_PORT = DEFAULT_DELEGATION_SERVER_PORT
+    
+    def local_client(self, **kwargs):
+        """
+        Returns a Client object to control the delegation server process itself.
+        
+        Args:
+            **kwargs: Additional arguments to override (e.g., sync, timeout)
+        
+        Returns:
+            Client: A Client object connected to the delegation server for direct control
+        """
+        return Client.from_client(self, **kwargs)
+
+    @property
+    def roster(self):
+        result = self.local_client().evaluate('roster')
+        if result.errored:
+            raise result.error
+        return result.value
+
+    @property
+    def delegator(self):
+        """ 
+        If you think self.evaluate_all is going to be slow because of bandwidth, consider self.delegator.evaluate_all
+        That way, all the workers will report directly to this client, instead of going through the self.delegator's server
+        """
+        result = self.local_client().evaluate('delegator')
+        if result.errored:
+            raise result.error
+        return result.value
+
     def evaluate(self, code, **vars):
         vars_string = ", ".join(x+'='+x for x in vars)
 
@@ -1270,6 +1407,193 @@ class DelegationClient(Client):
 
         return result
 
+    def evaluate_all(self, code, **vars):
+        """
+        Submits a job to be executed on all workers simultaneously via the delegation server.
+        
+        Args:
+            code (str): The Python code to execute on all workers.
+            **vars: Additional variables to pass to all workers.
+
+        Returns:
+            list[Evaluation]: Results from each worker in order.
+        """
+        vars_string = ", ".join(x+'='+x for x in vars)
+
+        delegation_code = rp.unindent(
+            """
+            %private_scope
+            %return result
+            import rp.web_evaluator as we
+            if 'delegator' not in dir():
+                raise we.NotADelegationServerError
+            result = delegator.evaluate_all(CODE, VARS)
+            """
+        ).replace('CODE', repr(code)).replace('VARS', vars_string)
+
+        result = super().evaluate(delegation_code, **vars)
+
+        if not result.errored:
+            result = result.value
+        else:
+            rp.fansi_print("Delegatee Error: "+str(result.error),'red','bold')
+
+        return result
+
+def launch_tmux_delegation_cluster(num_workers:int, *, session_name=None, 
+                                   worker_port=None, 
+                                   python_executable=None,
+                                   command_before=None,
+                                   attach=False,
+                                   roster=None,
+                                   delegator_port=None,
+                                   ip_address=None,
+                                   if_exists='error',
+                                   persistent_workers=False,
+                                   ):
+    """
+    Starts a tmux session with a cluster of web_evaluator servers for distributed computation.
+    This function makes life easy - it's intended to be general-purpose and feature-rich
+    (This is in contrast to the core web_evaluator Client/server protocol, which is intentionally minimalistic)
+
+    Usage notes:
+        To make debugging easier, each server comes with a rp.pterm() repl - so you can debug on the fly
+
+        This includes the delegator. So, if you want to test all clients for instance, you can run
+            delegator.evaluate_all("print(123)")
+        in the delegator window. 
+
+        Pro-tip: The worker servers are all in the same window - so you can also run commands in all of them with
+            tmux setw synchronize-panes
+    
+        The workers that are created have the following python variables:
+            WE_PORT : int #The port of this worker's we.run_server
+            WE_CLUSTER_RANK : int #If created by this function, is the index of the created worker
+
+            So, you can send commands accordingly - useful for things like GPU allocation etc.
+
+        Best practice: Send commands to workers that are self-contained and use singletons to cache any expensive operations
+            Reason why: new workers might be added on the fly, or rebooted - so sending commands that take care of the whole
+            setup process plus the part you want to run is an easy way to develop. Think importing a custom module - the code
+            inside it will only run once, and subsequent calls will be faster.
+
+    Args:
+        num_workers (int): Number of server instances to start
+        session_name (str, optional): Name for the tmux session. If None, generates unique name.
+        base_port (int, optional): Starting port number. Defaults to DEFAULT_SERVER_PORT.
+        python_executable (str, optional): Python executable to use. Defaults to sys.executable.
+        command_before (str/list, optional): Commands to run before starting servers (e.g., conda activate).
+        attach (bool, optional): Whether to attach to tmux session. Defaults to False.
+        roster (None/False/ClientRoster/str, optional): Roster handling:
+            - None: Use default roster
+            - False: Don't register clients to any roster  
+            - ClientRoster: Use this roster
+            - str: Evaluate string to get roster
+        delegator_port (int/None/False, optional): Port for delegation server:
+            - None: Auto-choose next available port from DEFAULT_DELEGATION_SERVER_PORT (default)
+            - False: Don't start delegation server
+            - int: Use exactly this port
+        ip_address (str, optional): IP address for clients. Defaults to local IP address.
+        if_exists (str, optional): What to do if session_name is a string, but already exists:
+            - 'error': Raise error if session exists (default)
+            - 'replace': Kill existing session and create new one
+            - 'unique': Create unique session name if name exists
+        persistent_workers (bool, optional): If True, if a worker dies, it will automatically restart
+            with the same port and WE_CLUSTER_RANK and WE_PORT as it used to have. This lets you do things
+            like call delegator.evaluate_all("rp.kill_process(rp.get_process_id())", sync=False) to reboot an entire set of workers
+    
+    Returns:
+        EasyDict: cluster info with 'session_name', 'clients', and optionally 'delegator'
+    
+    Example:
+        >>> cluster = launch_tmux_delegation_cluster(4, session_name="my_cluster")
+        >>> delegator = ClientDelegator(cluster['clients'])
+        >>> delegator.evaluate_all("print('Hello World!')")
+
+    """
+    if session_name is None:
+        session_name = rp.tmux_get_unique_session_name(DEFAULT_CLUSTER_SESSION_NAME)
+    if base_port is None:
+        base_port = DEFAULT_SERVER_PORT
+    if python_executable is None:
+        python_executable = sys.executable
+    if ip_address is None:
+        ip_address = rp.get_my_local_ip_address()
+    
+    # Handle session name conflicts
+    if rp.tmux_session_exists(session_name):
+        if if_exists == 'error':
+            raise ValueError(f"Tmux session '{session_name}' already exists. Use if_exists='replace' or 'unique'.")
+        elif if_exists == 'replace':
+            rp.tmux_kill_session(session_name)
+            rp.fansi_print(f"Killed existing session '{session_name}'", 'yellow')
+        elif if_exists == 'unique':
+            original_name = session_name
+            session_name = rp.tmux_get_unique_session_name(session_name)
+            rp.fansi_print(f"Session '{original_name}' exists, using '{session_name}' instead", 'yellow')
+        else:
+            raise ValueError(f"Invalid if_exists: {if_exists}. Must be 'error', 'replace', or 'unique'.")
+    
+    # Handle roster parameter
+    if roster is None:
+        roster = ClientRoster()
+    elif roster is False:
+        pass  # Don't register to any roster
+    elif isinstance(roster, ClientRoster):
+        pass  # Use provided roster
+    elif isinstance(roster, str):
+        roster = eval(roster)
+    else:
+        raise ValueError(f"Invalid roster type: {type(roster)}. Must be None, False, ClientRoster, or str.")
+    
+    # Generate ports for each server
+    ports = []
+    for i in range(num_workers):
+        port = rp.get_next_free_port(base_port, n=i)
+        if port != base_port + i:
+            rp.fansi_print(f"Warning: Port {base_port + i} taken, using {port}", 'yellow')
+        ports.append(port)
+    
+    # Create windows: servers in first window, delegation server in second if requested
+    # worker_commands = [f'{shlex.quote(python_executable)} -m rp.web_evaluator server -p {port}' for port in ports]
+    worker_commands = [f'''{shlex.quote(python_executable)} -c "WE_CLUSTER_RANK={rank}; WE_PORT={port}; import rp, rp.web_evaluator as we ; we.run_server({port}, new_thread=True) ; rp.pterm(globals(), level_title='Port={port}')"''' for rank, port in enumerate(ports)]
+    
+    if persistent_workers:
+        worker_commands = [f'while true ; do {command} ; done' for command in worker_commands]
+
+    windows = {"servers": worker_commands}
+    
+    if delegator_port is not False:
+        if delegator_port is None:
+            delegator_port = rp.get_next_free_port(DEFAULT_DELEGATION_SERVER_PORT)
+        elif isinstance(delegator_port, int):
+            if rp.get_port_is_taken(delegator_port):
+                raise ValueError(f"Delegator port {delegator_port} is already taken")
+        else:
+            raise ValueError(f"Invalid delegator_port: {delegator_port}. Must be None, False, or int.")
+        
+        delegation_cmd = f'''{shlex.quote(python_executable)} -c "import rp, rp.web_evaluator; rp.globalize_locals(rp.web_evaluator.run_delegation_server)({delegator_port}, new_thread=True) ; rp.pterm(globals(), level_title='Delegator Port={port}')"'''
+        windows["delegator"] = delegation_cmd
+    
+    yaml_config = rp.tmuxp_create_session_yaml(
+        windows, session_name=session_name, command_before=command_before
+    )
+    rp.tmuxp_launch_session_from_yaml(yaml_config, attach=attach)
+    
+    # Create and register clients
+    clients = [Client(ip_address, port) for port in ports]
+    
+    if roster is not False:
+        for client in clients:
+            roster.register(client.server_port, client.server_name, sync=client.sync)
+    
+    cluster_info = {'session_name': session_name, 'clients': clients}
+    if delegator_port is not False:
+        cluster_info['delegator'] = DelegationClient(ip_address, delegator_port)
+    
+    rp.fansi_print(f"Started cluster '{session_name}' with {len(ports)} servers", 'green', 'bold')
+    return rp.as_easydict(cluster_info)
+
 def interactive_mode():
     if rp.input_yes_no("Is this the server? No means this is the client."):
         print("Running the server.")
@@ -1289,40 +1613,38 @@ if __name__ == '__main__':
     try:
         parser = argparse.ArgumentParser(prog='web_evaluator.py', description='Web Evaluator: Remote Python code execution server and client',
                                          formatter_class=argparse.RawTextHelpFormatter)
-        parser.add_argument('-p', '--port', type=int, default=DEFAULT_SERVER_PORT, help='Server port (default: 43234)')
 
         subparsers = parser.add_subparsers(dest='mode', title='Modes', metavar='{server,s,client,c}', help='Select "server" or "client" mode')
 
         # Server mode arguments
         server_parser = subparsers.add_parser('server', aliases=['s'], help='Run the script as a server')
+        server_parser.add_argument('-p', '--port', type=int, default=DEFAULT_SERVER_PORT, help='Server port (default: 43234)')
 
         # Client mode arguments
         client_parser = subparsers.add_parser('client', aliases=['c'], help='Run the script as a client')
         client_parser.add_argument('server_ip', type=str, help='Server IP address (e.g., 192.168.1.10)')
+        client_parser.add_argument('-p', '--port', type=int, default=DEFAULT_SERVER_PORT, help='Server port (default: 43234)')
 
         parser.epilog = '''Examples:
-      Run the script interactively (no arguments):
-        python3 script_name.py
+      Run interactively (no arguments):
+        python -m rp.web_evaluator
 
       Run as a server with the default port:
-        python3 script_name.py server
+        python -m rp.web_evaluator server
 
       Run as a server with a custom port:
-        python3 script_name.py server -p 12345
+        python -m rp.web_evaluator server -p 12345
 
       Run as a client connecting to a server IP and default port:
-        python3 script_name.py client 192.168.1.10
+        python -m rp.web_evaluator client 192.168.1.10
 
       Run as a client connecting to a server IP and custom port:
-        python3 script_name.py client 192.168.1.10 -p 12345
+        python -m rp.web_evaluator client 192.168.1.10 -p 12345
         '''
 
-        args, unknown = parser.parse_known_args()
+        args = parser.parse_args()
 
-        if unknown:
-            print("Unknown arguments provided. Switching to interactive mode.")
-            interactive_mode()
-        elif args.mode in ('server', 's'):
+        if args.mode in ('server', 's'):
             print("Running the server.")
             run_server(server_port=args.port)
         elif args.mode in ('client', 'c'):
@@ -1337,5 +1659,3 @@ if __name__ == '__main__':
     except KeyboardInterrupt:
         print("Received KeyboardInterrupt. Exiting.")
 
-
-
